/**
 * Minimal state management.
 *
 * const actions = { inc: (atom, payload) => {} }
 * const atom = createAtom({ count: 1 }, actions)
 *
 * atom.observe(atom => console.log(atom.get()))
 *
 * atom.get() // { count: 1 }
 * atom.dispatch('increment') // action
 * atom.dispatch('increment', { by: 2 }) // action with payload
 */
module.exports = function createAtom (initialState, actions, options) {
  if ( initialState === void 0 ) initialState = {};
  if ( actions === void 0 ) actions = {};
  if ( options === void 0 ) options = {};

  var state = initialState
  var actionSeq = 0
  var listeners = []
  var debug = options.debug
  var get = function () { return state; }
  var set = createSet()
  var swap = createSwap()
  var dispatch = createDispatch()
  var bindActions = options.bindActions || defaultBindActions
  var atom = { get: get, set: set, swap: swap, dispatch: dispatch, observe: observe, fuse: fuse, actions: bindActions(dispatch, actions) }
  var evolve = options.evolve || defaultEvolve
  return atom

  function defaultEvolve (atom, action, actions) {
    if (!actions[action.type]) { throw new Error(("Missing action: " + (action.type))) }
    return actions[action.type](atom, action.payload)
  }

  function defaultBindActions (dispatch, actions) {
    return Object.keys(actions).reduce(function (boundActions, actionType) {
      boundActions[actionType] = function (payload) { return dispatch(actionType, payload); }
      return boundActions
    }, {})
  }

  function insert (listeners, f, i) {
    if (typeof i !== 'undefined') {
      for (var j = 0; j < listeners.length; j++) {
        if (listeners[j].i > i) {
          listeners.splice(j, 0, { f: f, i: i })
          return
        }
      }
    }
    listeners.push({ f: f, i: i })
  }

  function observe (f, i) {
    insert(listeners, f, i)
    return function unobserve () {
      for (var j = 0; j < listeners.length; j++) {
        if (listeners[j].f === f) {
          listeners.splice(j, 1)
          break
        }
      }
    }
  }

  function fuse (moreState, moreActions) {
    Object.assign(actions, moreActions)
    atom.actions = bindActions(dispatch, actions)
    if (moreState) { set(moreState) }
  }

  function createDispatch (sourceActions) {
    sourceActions = sourceActions || []
    return function dispatch (type, payload) {
      var action = { seq: ++actionSeq, type: type }
      if (typeof payload !== 'undefined') { action.payload = payload }
      if (debug) {
        report('action', action, sourceActions)
        var history = sourceActions.concat([action])
        var dispatch$1 = createDispatch(history)
        var set = createSet(history)
        var swap = createSwap(history)
        var boundActions = bindActions(dispatch$1, actions)
        var actionAtom = Object.assign({}, atom, { dispatch: dispatch$1, actions: boundActions, set: set, swap: swap })
        return evolve(actionAtom, action, actions)
      } else {
        return evolve(atom, action, actions)
      }
    }
  }

  function createSet (sourceActions, ref) {
    if ( ref === void 0 ) ref = {};
    var swap = ref.swap; if ( swap === void 0 ) swap = false;

    sourceActions = sourceActions || []
    return function set (update, options) {
      if ( options === void 0 ) options = {};

      var action = { payload: update }
      var prevState = state
      state = swap ? action.payload : Object.assign({}, state, action.payload)
      if (debug) { report('update', action, sourceActions, prevState) }
      listeners.forEach(function (l) { return l.f(atom); })
    }
  }

  function createSwap (sourceActions) {
    return createSet(sourceActions, { swap: true })
  }

  function report (type, action, sourceActions, prevState) {
    var info = { type: type, action: action, sourceActions: sourceActions, atom: atom }
    if (prevState) { info.prevState = prevState }
    typeof debug === 'function' ? debug(info) : debug.forEach(function (debug) { return debug(info); })
  }
}
