var computeDiff = require('deep-diff')

var dictionary = {
  N: { color: '#4CAF50', text: 'ADDED', atext: 'added' },
  E: { color: '#2196F3', text: 'UPDATED' },
  D: { color: '#F44336', text: 'DELETED', atext: 'deleted' },
  A: { color: '#2196F3', text: 'ARRAY' }
}

module.exports = function (options) {
  if ( options === void 0 ) options = {};

  options.diff = typeof options.diff === 'undefined' ? true : options.diff
  options.actions = typeof options.actions === 'undefined' ? false : options.actions
  options.updates = typeof options.updates === 'undefined' ? true : options.updates
  options.include = options.include || []
  options.exclude = options.exclude || []
  var logger = options.logger || console

  var tryCatch = function (fn, elseFn) {
    try {
      fn()
    } catch (e) {
      elseFn()
    }
  }
  var log = function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return logger.log.apply(logger, args);
  }
  var groupStart = function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return tryCatch(function () { return logger.groupCollapsed.apply(logger, args); }, function () { return logger.log.apply(logger, args); });
  }
  var groupEnd = function () { return tryCatch(function () { return logger.groupEnd(); }, function () {}); }

  return function (ref) {
    var type = ref.type;
    var atom = ref.atom;
    var action = ref.action;
    var sourceActions = ref.sourceActions;
    var prevState = ref.prevState;

    var sourceAction = type === 'action' ? action : sourceActions[sourceActions.length - 1] || {}
    if (options.include.length && !options.include.includes(sourceAction.type)) { return }
    if (options.exclude.length && options.exclude.includes(sourceAction.type)) { return }

    if (type === 'action' && options.actions) {
      var actions = sourceActions.concat(action)
      groupStart(("🚀 " + (actions.map(function (a) { return a.type; }).join(' → '))))
      log('payload', action.payload)
      log('chain', actions)
      groupEnd()
    }

    if (type === 'update' && options.updates) {
      var diff
      var diffSummaryMap = {}
      var diffSummary = []
      if (options.diff) {
        diff = computeDiff(prevState, atom.get())
        if (diff) {
          diff.forEach(function (change) {
            diffSummaryMap[change.kind] = diffSummaryMap[change.kind] ? diffSummaryMap[change.kind] + 1 : 1
          })
          Object.keys(dictionary).forEach(function (kind) {
            if (diffSummaryMap[kind]) {
              diffSummary.push(((diffSummaryMap[kind]) + " " + (dictionary[kind].text.toLowerCase())))
            }
          })
        }
      }

      groupStart(("🙌 " + (sourceActions.map(function (a) { return a.type; }).join(' → ') || '–') + " »"), diffSummary.join(', '))
      log('curr', atom.get())
      log('prev', prevState)
      log('update', action.payload)
      log.apply(void 0, [ 'chain' ].concat( sourceActions ))
      if (diff) { printDiff(diff) }
      groupEnd()
    }
  }

  function style (kind) {
    return ("color: " + (dictionary[kind].color) + "; font-weight: bold;")
  }

  function render (diff) {
    var kind = diff.kind;
    var path = diff.path;
    var lhs = diff.lhs;
    var rhs = diff.rhs;
    var index = diff.index;
    var item = diff.item;

    path = path || []

    switch (kind) {
      case 'E':
        return [path.join('.'), lhs, '→', rhs]
      case 'N':
        return [path.join('.'), rhs]
      case 'D':
        return [path.join('.')]
      case 'A':
        return [((path.join('.')) + "[" + index + "]"), dictionary[item.kind].atext].concat(render(item).slice(1))
      default:
        return []
    }
  }

  function printDiff (diff) {
    diff.forEach(function (elem) {
      var kind = elem.kind;
      logger.log.apply(logger, [ ("%c" + (dictionary[kind].text)), style(kind) ].concat( render(elem) ))
    })
  }
}
