var Preact = require('preact')
var raf = require('./raf')
var printDebug = require('./debug')

var dev = process.env.NODE_ENV !== 'production'
var canUseDOM = !!(typeof window !== 'undefined' && window.document && window.document.createElement)

function createContext () {
  var Provider = /*@__PURE__*/(function (superclass) {
    function Provider () {
      superclass.apply(this, arguments);
    }

    if ( superclass ) Provider.__proto__ = superclass;
    Provider.prototype = Object.create( superclass && superclass.prototype );
    Provider.prototype.constructor = Provider;

    Provider.prototype.getChildContext = function getChildContext () {
      return {
        atom: this.props.atom,
        debug: this.props.debug
      }
    };

    Provider.prototype.render = function render () {
      return this.props.children[0]
    };

    return Provider;
  }(Preact.Component));

  var Consumer = /*@__PURE__*/(function (superclass) {
    function Consumer (props, ref) {
      var this$1 = this;
      var atom = ref.atom;

      superclass.call(this)
      this.state = this.map(atom.get(), props)
      this.isPure = typeof props.pure === 'undefined' ? true : props.pure
      this.shouldObserve = typeof props.observe === 'undefined' ? canUseDOM : props.observe
      this.scheduleUpdate = props.sync ? function () { return this$1.update(); } : raf(function () { return this$1.update(); })
    }

    if ( superclass ) Consumer.__proto__ = superclass;
    Consumer.prototype = Object.create( superclass && superclass.prototype );
    Consumer.prototype.constructor = Consumer;

    Consumer.prototype.observe = function observe () {
      var this$1 = this;

      if (!this.shouldObserve) { return }
      var ref = this.context;
      var atom = ref.atom;
      this.unobserve && this.unobserve()
      this.unobserve = atom.observe(function () { this$1.cancelUpdate = this$1.scheduleUpdate() })
      this.observedAtom = atom
      delete this.boundActions
      delete this.boundActionsSpec
    };

    Consumer.prototype.componentWillUnmount = function componentWillUnmount () {
      this.unobserve && this.unobserve()
      this.cancelUpdate && this.cancelUpdate()
      delete this.unobserve
      delete this.cancelUpdate
      delete this.observedAtom
      delete this.boundActions
      delete this.boundActionsSpec
    };

    Consumer.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
      if (!this.isPure) { return true }

      // if it's <Consumer> with dynamic children, shortcut the check
      if (this.props.children !== nextProps.children) {
        if (dev && (this.context.debug || this.props.debug)) {
          printDebug.children(this.debugName(), this.props.children, nextProps.children)
        }
        return true
      }

      // our state is mappedProps, this is the main optimisation
      if (differ(this.state, nextState)) {
        if (dev && (this.context.debug || this.props.debug)) {
          printDebug.props(this.debugName(), this.state, nextState)
        }
        return true
      }

      // in connect() case don't need to diff further, no extra props
      if (this.props.originalProps) {
        this.cancelUpdate && this.cancelUpdate()
        return false
      }

      // in <Consumer /> case we also diff props
      if (differ(this.props, nextProps)) {
        if (dev && (this.context.debug || this.props.debug)) {
          printDebug.props(this.debugName(), this.props, nextProps)
        }
        return true
      }

      this.cancelUpdate && this.cancelUpdate()
    };

    Consumer.prototype.debugName = function debugName () {
      return this._component
        ? this._component.constructor.name
        : this._parentComponent.constructor.name + ' → ' + this.constructor.name
    };

    Consumer.prototype.componentWillReceiveProps = function componentWillReceiveProps (nextProps) {
      var ref = this.context;
      var atom = ref.atom;
      var nextMappedProps = this.map(atom.get(), nextProps)
      this.setState(nextMappedProps)
    };

    Consumer.prototype.componentDidUpdate = function componentDidUpdate () {
      this.cancelUpdate && this.cancelUpdate()
    };

    Consumer.prototype.map = function map (state, props) {
      var originalProps = props.originalProps;
      var map = props.map;
      return Object.assign({}, originalProps, map ? map(state, originalProps) : {})
    };

    Consumer.prototype.update = function update () {
      var ref = this.context;
      var atom = ref.atom;
      var nextMappedProps = this.map(atom.get(), this.props)
      this.setState(nextMappedProps)
    };

    Consumer.prototype.bindActions = function bindActions (actions, dispatch, mappedProps) {
      if (!actions) { return { dispatch: dispatch } }
      if (typeof actions === 'function') { return actions(dispatch, mappedProps) }
      if (!this.boundActions || this.boundActionsSpec !== actions) {
        this.boundActionsSpec = actions
        this.boundActions = actions.reduce(function (acc, action) {
          acc[action] = function (payload) { return dispatch(action, payload); }
          return acc
        }, {})
      }
      return this.boundActions
    };

    Consumer.prototype.render = function render (ref, state, ref$1) {
      var actions = ref.actions;
      var originalProps = ref.originalProps;
      var render = ref.render;
      var children = ref.children;
      var atom = ref$1.atom;

      // do this in render, because:
      //  doing in constructor would cause memory leaks in SSR
      //  doing in componentDidMount leads to the wrong order of subscriptions
      //  we don't have another hook in Preact's context to check if atom was swapped
      if (!this.unobserve || this.observedAtom !== this.context.atom) { this.observe() }
      var mappedProps = this.state
      var boundActions = this.bindActions(actions, atom.dispatch, mappedProps)
      return (render || children[0])(Object.assign({}, originalProps, mappedProps, boundActions))
    };

    return Consumer;
  }(Preact.Component));

  function connect (map, actions, options) {
    if ( options === void 0 ) options = {};

    return function connectComponent (Component) {
      var render = function (mappedProps) { return Preact.h( Component, mappedProps); }
      var Connected = function (props) { return (
        Preact.h( Consumer, {
          map: map, actions: actions, pure: options.pure, sync: options.sync, observe: options.observe, debug: options.debug, originalProps: props, render: render })
      ); }
      return Connected
    }
  }

  function differ (mappedProps, nextMappedProps) {
    for (var i in mappedProps) {
      if (mappedProps[i] !== nextMappedProps[i]) { return true }
    }
    for (var i$1 in nextMappedProps) {
      if (!(i$1 in mappedProps)) { return true }
    }
    return false
  }

  return { Provider: Provider, Consumer: Consumer, connect: connect }
}

var ref = createContext();
var Provider = ref.Provider;
var Consumer = ref.Consumer;
var connect = ref.connect;
module.exports = { Provider: Provider, Consumer: Consumer, connect: connect, createContext: createContext }
